@@ -69,16 +69,12 @@
 #include "xf86Resources.h"
 #include "xf86RAC.h"
 
-#include "xf1bpp.h"
-#include "xf4bpp.h"
 #include "fb.h"
 
-#ifdef USE_AFB
-#include "afb.h"
+#ifdef XSERVER_LIBPCIACCESS
+#include <pciaccess.h>
 #endif
 
-#include "mfb.h"
-
 #define VESA_VERSION		4000
 #define VESA_NAME		"VESA"
 #define VESA_DRIVER_NAME	"vesa"
@@ -95,8 +91,12 @@ typedef struct _VESARec
     CARD16 major, minor;
     VbeInfoBlock *vbeInfo;
     GDevPtr device;
+#ifdef XSERVER_LIBPCIACCESS
+    struct pci_device *pciInfo;
+#else
     pciVideoPtr pciInfo;
     PCITAG pciTag;
+#endif
     miBankInfoRec bank;
     int curBank, bankSwitchWindowB;
     CARD16 maxBytesPerScanline;
@@ -108,7 +108,7 @@ typedef struct _VESARec
     CARD32 *pal, *savedPal;
     CARD8 *fonts;
     xf86MonPtr monitor;
-    Bool shadowFB, primary;
+    Bool shadowFB, strict_validation;
     CARD32 windowAoffset;
     /* Don't override the default refresh rate. */
     Bool defaultRefresh;
@@ -117,6 +117,8 @@ typedef struct _VESARec
     int nDGAMode;
     CloseScreenProcPtr CloseScreen;
     CreateScreenResourcesProcPtr CreateScreenResources;
+    xf86EnableDisableFBAccessProc *EnableDisableFBAccess;
+    Bool accessEnabled;
     OptionInfoPtr Options;
     IOADDRESS ioBase;
     Bool ModeSetClearScreen;
